import torch
from typing import Optional, Dict
import logging
from torch.utils.data import DataLoader
from diresa_torch.arch.models import Diresa
from operator import add

# TODO: Could improve visual feedback for training use tqdm


def __compute_losses(outputs, targets, criterions, loss_weights):
    """
    Helper function to compute losses given outputs, targets, criterions and weights.

    :param outputs: Model outputs (tuple of 3 elements: reconstructed, latent, distance)
    :param targets: Target values (tuple of 3 elements: data, None, None)
    :param criterions: List of loss functions [ReconstructionLoss, CovarianceLoss, DistanceLoss]
    :param loss_weights: Weighting factor for the different losses
    :return: Tuple of (individual_losses, total_weighted_loss)
    """
    individual_losses = [c(o, t) for c, o, t in zip(criterions, outputs, targets)]
    weighted_losses = [w * l for w, l in zip(loss_weights, individual_losses)]
    total_weighted_loss = torch.stack(weighted_losses).sum()

    return individual_losses, total_weighted_loss


def __set_non_trainable(model):
    for param in model.parameters():
        param.requires_grad = False


def __evaluate(
    produce_output: callable,
    produce_target: callable,
    test_loader: DataLoader,
    criterions: list,
    device: str = 'cpu',
    loss_weights: list = [1., 1., 1.]
) -> Dict[str, float]:
    """
    Evaluate DIRESA (does not track gradient) by computing all three losses (reconstruction, covariance, distance)
    with help of the ``produce_output`` and ``produce_input`` functions. Those functions are provided as lambdas
    which makes it easier to match the outputs, targets and criterions together for evaluation purposes.

    :param produce_output: callable function producing outputs from model. Takes as input batch data.
    :param produce_target: callable function producing target values for criterion. Takes as input batch data.
    :param test_loader: Test data loader
    :param criterions: List of loss functions, depends on what part is being evaluated.
    :param device: Device to evaluate on
    :param loss_weights: Weighting factor for the different losses

    :return: Dictionary with average losses: individual losses + weighted total loss
    """
    total_losses = [0.0] * (len(criterions) + 1)
    num_batches = 0

    loss_names = [f"{c}"[:-2] for c in criterions] + ['WeightedLoss']

    with torch.no_grad():
        for _, data in enumerate(test_loader):
            data = data.to(device)

            outputs = produce_output(data)

            targets = produce_target(data)

            # outputs_losses, loss = _compute_losses(outputs, target, criterions, loss_weights)
            outputs_losses, loss = __compute_losses(
                outputs, targets, criterions, loss_weights)

            # Accumulate losses
            all_losses = outputs_losses + [loss]
            total_losses = list(
                map(add, total_losses, [loss.item() for loss in all_losses]))
            num_batches += 1

    avg_losses = [loss / num_batches for loss in total_losses]

    result = {name: avg_loss for name, avg_loss in zip(loss_names, avg_losses)}

    return result


def train_diresa(
    model: Diresa,
    train_loader: DataLoader,
    criterions: list,
    optimizer: torch.optim.Optimizer,
    num_epochs: int = 10,
    device: str = 'cpu',
    val_loader: Optional[DataLoader] = None,
    callbacks: Optional[list] = None,  # Not IMPL at the moment -> see assert
    loss_weights: list = [1., 1., 1.],
    staged_training: bool = False
) -> Dict[str, list]:
    """
        Trains DIRESA.

        :param model: The model to train
        :param train_loader: Training data loader
        :param criterion: List of Loss function. With order [ReconstructionLoss, CovarianceLoss, DistanceLoss]
        :param optimizer: Optimizer
        :param num_epochs: Number of epochs
        :param device: Device to train on
        :param val_loader: Optional validation loader
        :param callbacks: Optional list of callback functions
        :param loss_weights: Weighting factor for the different losses.
        :param staged_trainig: If set to True will train the encoder and the decoder separately for ``num_epochs`` each.

        :return Dict with training (losses, metrics) and validation (if val_loader is provided).
    """
    assert callbacks is None, "Callbacks are not tested at the moment. Remove param or set to None"

    def __train_for_epochs(produce_output: callable, produce_target: callable, criterions, loss_weights, prepend_log="DIRESA"):
        """
            Nested function for training loop. Factors out common functionalities for training,
            while providing custom informations about what loss to train for used to differentiate
            between staged training and full training.

            :param produce_output: callable function producing outputs from model. Takes as input batch data.
            :param produce_target: callable function producing target values for criterion. Takes as input batch data.
            :param criterions: List of loss functions
            :param loss_weights: weights for each loss function
            :param prepend_log: String to prepend to logging output
        """
        assert len(criterions) == len(
            loss_weights), "Number of criterions and their associated weights does not match"

        # Ordering of loss ouput values
        # Use class name without the last () as name for the loss
        loss_names = [f"{c}"[:-2] for c in criterions] + ['WeightedLoss']

        # add "train" suffix for output
        train_loss_names = list(map(lambda x: x + "_train", loss_names))
        history = {loss_name: [] for loss_name in train_loss_names}

        for epoch in range(num_epochs):
            # each criterion loss + combined weighted loss
            epoch_loss = [0.0] * (len(criterions) + 1)
            num_batches = 0

            for batch_idx, data in enumerate(train_loader):

                data = data.to(device)
                target = produce_target(data)
                outputs = produce_output(data)

                optimizer.zero_grad()
                outputs_losses, loss = __compute_losses(
                    outputs, target, criterions, loss_weights)

                # accumulates gradient in each tensor -> Backprop
                # backpropagated loss in weighted sum of each loss.
                loss.backward()

                optimizer.step()

                # add weighted loss to final losses
                all_losses = outputs_losses + [loss]
                epoch_loss = list(map(add, epoch_loss, [loss.item() for loss in all_losses]))
                num_batches += 1

                if callbacks:
                    for callback in callbacks:
                        callback(epoch, batch_idx, loss.item())

            avg_loss = list(map(lambda loss: loss / num_batches, epoch_loss))

            for k, value in zip(train_loss_names, avg_loss):
                history[k].append(value)

            # val loader is defined in exterior function
            if val_loader:
                val_dict = __evaluate(
                    produce_output,
                    produce_target,
                    val_loader,
                    criterions,
                    device,
                    loss_weights
                )
                renamed_val = {f"{k}_val": v for k, v in val_dict.items()}
                for k, value in renamed_val.items():
                    if k in history:
                        history[k].append(value)
                    else:
                        history[k] = [value]

            # print out last entry in history for each epoch
            log_str = ", ".join(
                [f"{name}: {values[-1]:.4e}" for name, values in history.items()])
            logging.info(
                f'{prepend_log}: Epoch {epoch + 1}/{num_epochs} - {log_str}')

        return history

    # End of nested function

    model: Diresa = model.to(device)

    if staged_training:
        # train encoder, cov and dist loss
        hist_encoder = __train_for_epochs(
            lambda data: model._encode_with_distance(data.x),
            lambda data: (None, None),
            criterions[1:],  # cov and dist criterions
            loss_weights[1:],  # cov and dist weights
            prepend_log="Encoder"
        )

        # freeze encoder weights
        __set_non_trainable(model.base_encoder)

        # train decoder, only rec loss
        hist_decoder = __train_for_epochs(
            lambda data: model.base_decoder(model.base_encoder(data.x)),
            lambda data: data.y,
            criterions[:1],
            loss_weights[:1],
            prepend_log="Decoder"
        )

        hist = {**hist_encoder, **hist_decoder}
        return hist

    else:
        # data is produced by forward pass of model.
        hist = __train_for_epochs(
            lambda data: model(data.x),
            lambda data: (data.y, None, None),
            criterions,
            loss_weights
        )

        return hist


def evaluate_diresa(
    model: Diresa,
    test_loader: DataLoader,
    criterions: list,
    device: str = 'cpu',
    loss_weights: list = [1., 1., 1.]
) -> Dict[str, float]:

    assert len(criterions) == 3, "Need to provide 3 criterions for DIRESA evaluation, namely [DistanceLoss, CovarianceLoss, DistanceLoss]"

    eval_dict = __evaluate(
        # model.forward(data) produces (reconstruced, latent, dist)
        lambda data: model.forward(data.x),
        lambda data: (data.y, None, None),
        test_loader=test_loader,
        criterions=criterions,
        device=device,
        loss_weights=loss_weights
    )
    log_str = ", ".join(
        [f"{name}_eval: {value:.4e}" for name, value in eval_dict.items()])
    logging.info(log_str)
    return eval_dict


def predict_diresa(
    model: Diresa,
    data_loader: DataLoader,
    device: str = 'cpu'
) -> torch.Tensor:
    """
    Generate predictions for DIRESA using a trained model
    Provides faster inference as it does not compute distance or covariance losses.
    """
    model: Diresa = model.to(device)

    predictions = []

    with torch.no_grad():
        for _, data in enumerate(data_loader):
            data = data.to(device)
            outputs = model.fast_eval(data)
            predictions.append(outputs.cpu())

    return torch.cat(predictions, dim=0)



import torch
import torch.nn as nn


class LatentCovLoss(nn.Module):
    """
    Latent covariance loss class
    """
    # TODO: Simmulated Annealing. Not present by default in pytorch. Needs custom implentation
    # With staged training might not really need it anymore as parameters are tuned independently

    def __init__(self):
        super().__init__()

    def forward(self, input, _):
        """
        :param  input: is the latent space representation of the current batch
        """
        cov = torch.cov(torch.transpose(input, 0, 1), correction=1)
        cov_square = cov * cov  # elem wise mult
        # number of covariance entries (need to subract 1 to not take into account diagonal variance metrics)
        nbr_of_cov = input.shape[-1] * (input.shape[-1] - 1)
        cov_loss = (torch.sum(cov_square) -
                    torch.trace(cov_square)) / float(nbr_of_cov)
        return cov_loss


class MAEDistLoss(nn.Module):
    """
    Absolute Error between original and latent distances

    :param distances: batch of original and latent distances between twins
    :return: batch of absolute errors
    """

    def __init__(self):
        super().__init__()

    def forward(self, input, _):
        distances = input
        ae = torch.abs(distances[:, 0] - distances[:, 1])
        return torch.mean(ae)


class MALEDistLoss(nn.Module):
    """
    Absolute Error between logarithm of original and latent distances

    :param distances: batch of original and latent distances between twins
    :return: batch of absolute logarithmic errors
    """

    def __init__(self):
        super().__init__()

    def forward(self, input, _):
        distances = input
        ale = torch.abs(torch.log1p(
            distances[:, 0]) - torch.log1p(distances[:, 1]))
        return torch.mean(ale)


class MAPEDistLoss(nn.Module):
    """
    Absolute Percentage Error between original and latent distances

    :param distances: batch of original and latent distances between twins
    :return: batch of absolute percentage errors
    """

    def __init__(self):
        super().__init__()

    def forward(self, input, _):
        distances = input
        epsilon = 1e-8
        ape = torch.abs(
            (distances[:, 0] - distances[:, 1]) / (distances[:, 0] + epsilon))
        return torch.mean(ape)


class MSEDistLoss(nn.Module):
    """
    Squared Error between original and latent distances

    :param distances: batch of original and latent distances between twins
    :return: batch of squared errors
    """

    def __init__(self):
        super().__init__()

    def forward(self, input, _):
        distances = input
        se = torch.square(distances[:, 0] - distances[:, 1])
        return torch.mean(se)


class MSLEDistLoss(nn.Module):
    """
    Squared Error between logarithm of original and latent distances

    :param distances: batch of original and latent distances between twins
    :return: batch of squared logarithmic errors
    """

    def __init__(self):
        super().__init__()

    def forward(self, input, _):
        distances = input
        sle = torch.square(torch.log1p(
            distances[:, 0]) - torch.log1p(distances[:, 1]))
        return torch.mean(sle)


class CorrDistLoss(nn.Module):
    """
    Correlation loss between original and latent distances

    :param distances: batch of original and latent distances between twins
    :return: 1 - correlation coefficient
    """

    def __init__(self):
        super().__init__()

    def forward(self, input, _):
        distances = input

        # Calculate covariance matrix
        x = distances[:, 0]
        y = distances[:, 1]

        # Center the data
        x_centered = x - torch.mean(x)
        y_centered = y - torch.mean(y)

        # Calculate covariance and variances
        cov_xy = torch.mean(x_centered * y_centered)
        var_x = torch.mean(x_centered * x_centered)
        var_y = torch.mean(y_centered * y_centered)

        # Calculate correlation coefficient
        std_x = torch.sqrt(torch.abs(var_x))
        std_y = torch.sqrt(torch.abs(var_y))

        # Avoid division by zero
        epsilon = 1e-8
        corr = cov_xy / (std_x * std_y + epsilon)

        return 1 - corr



import torch
from typing import Literal, Optional, Dict, overload
import logging
from torch.utils.data import DataLoader
from diresa_torch.arch.models import Diresa
from operator import add
from copy import deepcopy


def __compute_losses(outputs, targets, criteria, loss_weights):
    """
    Helper function to compute losses given outputs, targets, criteria and weights.

    :param outputs: Model outputs (tuple of 3 elements: reconstructed, latent, distance)
    :param targets: Target values (tuple of 3 elements: data, None, None)
    :param criteria: List of loss functions [ReconstructionLoss, CovarianceLoss, DistanceLoss]
    :param loss_weights: Weighting factor for the different losses
    :return: Tuple of (individual_losses, total_weighted_loss)
    """
    individual_losses = [c(o, t) for c, o, t in zip(criteria, outputs, targets)]
    weighted_losses = [w * l for w, l in zip(loss_weights, individual_losses)]
    total_weighted_loss = torch.stack(weighted_losses).sum()

    return individual_losses, total_weighted_loss


def __set_non_trainable(model):
    for param in model.parameters():
        param.requires_grad = False


def __loss_string_repr(criteria, suffix=""):
    """
    Helper functions producing string criteria of
    losses. String repr is used to log each loss with an
    understandable name.
    """
    # Ordering of loss ouput values
    # Use class name without the last () as name for the loss
    loss_names = [f"{c}"[:-2] for c in criteria] + ["WeightedLoss"]

    # Criteria list can change depending on training mode.
    # When runing normal criteria[0] is Reconstruction loss (and len(criteria == 3))
    # During staged training, when encoder is trained
    # criteria[0] is LatentCovLoss (and len(criteria == 2)).
    # When decoder is is trained criteria[0] is Recon (and len(criteria == 1))
    if len(criteria) == 3 or len(criteria) == 1:
        loss_names[0] = "Recon" + loss_names[0]

    # add "train" suffix for output
    loss_names = list(map(lambda x: x + suffix, loss_names))
    return loss_names


def __evaluate(
    produce_output: callable,
    produce_target: callable,
    test_loader: DataLoader,
    criteria: list,
    device: str = "cpu",
    loss_weights: list = [1.0, 1.0, 1.0],
    loss_suffix: str = "_eval",
) -> Dict[str, float]:
    """
    Evaluate DIRESA (does not track gradient) by computing all three losses (reconstruction, covariance, distance) with help of the ``produce_output`` and ``produce_input`` functions. Those functions are provided as lambdas which makes it easier to match the outputs, targets and criteria together for evaluation purposes.

    :param produce_output: callable function producing outputs from model. Takes as input batch data.
    :param produce_target: callable function producing target values for criterion. Takes as input batch data.
    :param test_loader: Test data loader
    :param criteria: List of loss functions, depends on what part is being evaluated.
    :param device: Device to evaluate on
    :param loss_weights: Weighting factor for the different losses
    :param suffix: Appends a suffix to the loss as __eval can be used for validation during training or evaluation (when passing in test set)

    :return: Dictionary with average losses: individual losses + weighted total loss
    """
    total_losses = [0.0] * (len(criteria) + 1)
    num_batches = 0

    loss_names = __loss_string_repr(criteria, loss_suffix)

    with torch.no_grad():
        for _, data in enumerate(test_loader):
            data = data.to(device)

            outputs = produce_output(data)

            targets = produce_target(data)

            # outputs_losses, loss = _compute_losses(outputs, target, criteria, loss_weights)
            outputs_losses, loss = __compute_losses(
                outputs, targets, criteria, loss_weights
            )

            # Accumulate losses
            all_losses = outputs_losses + [loss]
            total_losses = list(
                map(add, total_losses, [loss.item() for loss in all_losses])
            )
            num_batches += 1

    avg_losses = [loss / num_batches for loss in total_losses]

    result = {name: avg_loss for name, avg_loss in zip(loss_names, avg_losses)}

    return result


def train_diresa(
    model: Diresa,
    train_loader: DataLoader,
    criteria: list,
    optimizer: torch.optim.Optimizer,
    num_epochs: int = 10,
    device: str = "cpu",
    val_loader: Optional[DataLoader] = None,
    callbacks: Optional[list] = None,  # Not IMPL at the moment
    loss_weights: list = [1.0, 1.0, 1.0],
    staged_training: bool = False,
) -> dict[str, dict[str, list[float]]]:
    """
    Trains DIRESA.

    :param model: The model to train
    :param train_loader: Training data loader
    :param criterion: List of Loss function. With order [ReconstructionLoss, CovarianceLoss, DistanceLoss]
    :param optimizer: Optimizer
    :param num_epochs: Number of epochs
    :param device: Device to train on
    :param val_loader: Optional validation loader
    :param callbacks: Optional list of callback functions
    :param loss_weights: Weighting factor for the different losses.
    :param staged_trainig: If set to True will train the encoder and the decoder separately for ``num_epochs`` each.

    :return Dict with training (losses, metrics) and validation (if val_loader is provided).
    """
    assert (
        callbacks is None
    ), "Callbacks are not tested at the moment. Remove param or set to None"

    def __train_for_epochs(
        produce_output: callable,
        produce_target: callable,
        criteria,
        loss_weights,
        prepend_log="DIRESA",
    ):
        """
        Nested function for training loop. Factors out common functionalities for training,
        while providing custom informations about what loss to train for used to differentiate
        between staged training and full training.

        :param produce_output: callable function producing outputs from model. Takes as input batch data.
        :param produce_target: callable function producing target values for criterion. Takes as input batch data.
        :param criteria: List of loss functions
        :param loss_weights: weights for each loss function
        :param prepend_log: String to prepend to logging output
        """
        assert len(criteria) == len(
            loss_weights
        ), "Number of criteria and their associated weights does not match"

        loss_names = __loss_string_repr(criteria, "_train")
        history = {name: [] for name in loss_names}

        for epoch in range(num_epochs):
            # each criterion loss + combined weighted loss
            epoch_loss = [0.0] * (len(criteria) + 1)
            num_batches = 0

            for _, data in enumerate(train_loader):
                data = data.to(device)
                target = produce_target(data)
                outputs = produce_output(data)

                optimizer.zero_grad()
                outputs_losses, loss = __compute_losses(
                    outputs, target, criteria, loss_weights
                )

                # accumulates gradient in each tensor -> Backprop
                # backpropagated loss in weighted sum of each loss.
                loss.backward()

                optimizer.step()

                # add weighted loss to final losses
                all_losses = outputs_losses + [loss]
                epoch_loss = list(
                    map(add, epoch_loss, [loss.item() for loss in all_losses])
                )
                num_batches += 1

                if callbacks:
                    raise NotImplementedError
                    # for callback in callbacks:
                    #     callback(epoch, batch_idx, loss.item())

            avg_loss = list(map(lambda loss: loss / num_batches, epoch_loss))

            for name, loss in zip(loss_names, avg_loss):
                history[name].append(loss)

            # val loader is defined in exterior function
            if val_loader:
                val_dict = __evaluate(
                    produce_output,
                    produce_target,
                    val_loader,
                    criteria,
                    device,
                    loss_weights,
                    loss_suffix="_val",
                )
                for name, loss in val_dict.items():
                    if name in history:
                        history[name].append(loss)
                    else:
                        history[name] = [loss]

            # print out last entry in history for each epoch
            log_str = ", ".join(
                [f"{name}: {values[-1]:.4e}" for name, values in history.items()]
            )
            logging.info(f"{prepend_log}: Epoch {epoch + 1}/{num_epochs} - {log_str}")

        return history

    # End of nested function

    model: Diresa = model.to(device)

    if staged_training:
        # train encoder, cov and dist loss
        hist_encoder = __train_for_epochs(
            lambda data: model._encode_with_distance(data),
            lambda data: (None, None),
            criteria[1:],  # cov and dist criteria
            loss_weights[1:],  # cov and dist weights
            prepend_log="Encoder",
        )

        # freeze encoder weights
        __set_non_trainable(model.base_encoder)

        # train decoder, only rec loss
        hist_decoder = __train_for_epochs(
            lambda data: model.base_decoder(model.base_encoder(data)),
            lambda data: data,
            criteria[:1],
            loss_weights[:1],
            prepend_log="Decoder",
        )

        hist = {"Encoder": hist_encoder, "Decoder": hist_decoder}
        return hist

    else:
        # data is produced by forward pass of model.
        hist_diresa = __train_for_epochs(
            lambda data: model(data),
            lambda data: (data, None, None),
            criteria,
            loss_weights,
            prepend_log="Encoder_Decoder",
        )

        # To keep consitency with staged training where encoder
        # and decoder losses are accessed via a key, the same is done
        # when all weights are trained simultaneously.
        hist = {"Encoder_Decoder": hist_diresa}
        return hist


def evaluate_diresa(
    model: Diresa,
    test_loader: DataLoader,
    critera: list,
    device: str = "cpu",
    loss_weights: list = [1.0, 1.0, 1.0],
) -> Dict[str, float]:
    assert (
        len(critera) == 3
    ), "Need to provide 3 criteria for DIRESA evaluation, namely [DistanceLoss, CovarianceLoss, DistanceLoss]"

    eval_dict = __evaluate(
        # model.forward(data) produces (reconstruced, latent, dist)
        lambda data: model.forward(data),
        lambda data: (data, None, None),
        test_loader=test_loader,
        criteria=critera,
        device=device,
        loss_weights=loss_weights,
        loss_suffix="_eval",
    )
    log_str = ", ".join([f"{name}: {value:.4e}" for name, value in eval_dict.items()])
    logging.info(log_str)
    return eval_dict


def predict_diresa(
    model: Diresa, data_loader: DataLoader, device: str = "cpu"
) -> torch.Tensor:
    """
    Generate predictions for DIRESA using a trained model
    Provides faster inference as it does not compute distance or covariance losses.
    """
    model: Diresa = model.to(device)

    predictions = []

    with torch.no_grad():
        for _, data in enumerate(data_loader):
            data = data.to(device)
            outputs = model.fast_eval(data)
            predictions.append(outputs.cpu())

    return torch.cat(predictions, dim=0)


def __set_components_to_mean(latent: torch.Tensor, retain_idx: int):
    """
    Sets all latent components to mean except the ones in the list (which are kept untouched)

    :param latent: latent dataset
    :param retain: components not in this list are set to mean
    :return: latent dataset with all components set to mean except the ones in the list
    """

    with torch.no_grad():
        mean_values = latent.mean(dim=0, keepdim=True)
        mask = torch.tensor(
            [i != retain_idx for i in range(latent.shape[1])],
            dtype=torch.bool,
            device=latent.device,
        )
        result = torch.where(mask, mean_values, latent)
    return result


def __r2_score(y: torch.Tensor, y_pred: torch.Tensor):
    """
    :param y: original dataset
    :param y_pred: predicted dataset
    :return: R2 score between y and y_pred
    """
    error = torch.sum(torch.square(y - y_pred))
    var = torch.sum(torch.square(y - torch.mean(y, dim=0)))
    r2 = 1.0 - error / var
    return r2.item()  # Convert to Python scalar


# As order_diresa either returns an new object or None, in order
# to help the type checker we provide type hints.
# TODO: Should move this to another file called types.py
@overload
def order_diresa(
    model: Diresa, data_loader: DataLoader, *, inplace: Literal[True]
) -> None: ...


# need one overload for when we use the default (inplace is not provided meaning it is set to true)
@overload
def order_diresa(model: Diresa, data_loader: DataLoader) -> None: ...


@overload
def order_diresa(
    model: Diresa, data_loader: DataLoader, *, inplace: Literal[False]
) -> Diresa: ...


# NOTE: Could rewrite all map functions using tensor operations or numpy instead of python builtin sequence operations
# NOTE: Return type is set to None when ordering change is made in place (i.e. changed
# directly in the model passed as reference).
def order_diresa(
    model: Diresa, data_loader: DataLoader, device: str = "cpu", inplace=True
) -> Diresa | None:
    """
    Sets ordering of the OrderingLayer.
    Limitations: assumes a flat latent space (rank of latent is 2)

    By default will set ordering in place in the provided model and return None.
    If `inplace` is set to false will return a new model for which the ordering has been set while leaving the original model passed as a parameter untouched.


    :param model: The model on which to produce the ordering
    :param data_loader: The data_loader from which to produce the ordering.
    :param device:
    :param inplace: If true returns a copy of the diresa object with new ordering,
    leaving the passed
    """
    r2_scores = []

    logging.info(f"Batch size for ordering is {data_loader.batch_size}")

    with torch.no_grad():
        # produce r2 per batch
        for _, data in enumerate(data_loader):
            data = data.to(device)
            # NOTE: Might need to encapsulate ordering in
            # a function working by batch for testing.
            latent = model.base_encoder(data)
            assert len(latent.shape) == 2, "Latent space is not flattend"

            # 2. Produce l latent samples for wich every latent dimensions is averaged except the l-th one.
            averaged = map(
                lambda i: __set_components_to_mean(latent, i),
                range(latent.shape[1]),
            )

            # 3. Produce l decoded samples from l latent samples
            decoded = map(lambda latent: model.base_decoder(latent), averaged)

            # 4. Compute R2 by comparing l decoded with orginal x
            r2 = list(map(lambda pred: __r2_score(data, pred), decoded))

            r2_scores.append(r2)

        r2_scores = torch.tensor(r2_scores).mean(dim=0)

        logging.info(
            f"Ordered R2 scores are: {list(map(lambda x: x.item(), sorted(r2_scores, reverse=True)))}"
        )

        ordering = torch.argsort(r2_scores, descending=True)

        if inplace:
            model.ordering_layer.order = ordering
            return None
        else:
            new_model: Diresa = deepcopy(model)
            new_model.ordering_layer.order = ordering
            return new_model



import torch
import torch.nn as nn

from diresa_torch.arch.modules import DenseLayer, DistanceLayer, CNNLayer, OrderingLayer
from diresa_torch.arch.config import DiresaSetup, DiresaConfig
from typing import Tuple, Optional
from functools import reduce
import operator

import logging


def cnn_output_flat(nb_stack, dims, cnn_output_filter):
    """
    Computes intermediate dimensions from CNN to FNN.
    Output dimensions of CNN is ``(original_x / 2^len(stack)) * (orignal_y / 2^len(stack)) * output filter``.
    :param nb_stack: number of stacks in CNN models. Len of nb_stack defines how many times the input dimensions are divided by 2 due to the MaxPool2d layer.
    :param dims: original input dimensions
    :param cnn_output_filter: filters of the last CNN layer.
    """
    spatial_dims = [dim // (2**nb_stack) for dim in dims]
    return reduce(operator.mul, spatial_dims + [cnn_output_filter])


class Encoder(nn.Module):
    """
    Provides the encoder for DIRESA.

    :param config: `DiresaConfig` object built by the static builder method DiresaSetup.create_config()
    """

    def __init__(self, config: DiresaConfig):
        super().__init__()

        if config.dense_config and not config.cnn_config:
            # NOTE:
            # Input of dense_units is input_shape[0].
            # By passing input dimension (config.input_shape[0])
            # by copy here we do not modify the original input_shape
            # If we mutate config.input_shape here in place we need to remember to NOT
            # modify it a second time in the decoder otherwise 2 layers are added.
            dense_units = (config.input_shape[0],) + config.dense_config.dense_units
            self.network = DenseLayer(dense_units, config.activation)

        elif config.cnn_config and not config.dense_config:
            self.cnn = CNNLayer(
                config.cnn_config.stack,
                config.cnn_config.stack_filters,
                config.cnn_config.kernel_size,
                config.input_shape,
                config.activation,
            )

            self.flatten = nn.Flatten()

            self.network = nn.Sequential(
                self.cnn,
                self.flatten,
            )

        elif config.cnn_config and config.dense_config:
            self.cnn = CNNLayer(
                config.cnn_config.stack,
                config.cnn_config.stack_filters,
                config.cnn_config.kernel_size,
                config.input_shape,
                config.activation,
            )

            ffn_input = cnn_output_flat(
                len(config.cnn_config.stack),
                config.input_shape[1:],
                config.cnn_config.stack_filters[-1],
            )

            dense_units = (ffn_input,) + config.dense_config.dense_units

            self.ffn = DenseLayer(dense_units, config.activation)

            self.flatten = nn.Flatten()

            self.network = nn.Sequential(self.cnn, self.flatten, self.ffn)

    def forward(self, x):
        return self.network(x)


class Decoder(nn.Module):
    def __init__(self, config: DiresaConfig):
        super().__init__()

        if config.dense_config and not config.cnn_config:
            # NOTE: Do not change config.input_shape in place, see comment in Encoder
            # Case where we only have dense units. Input of dense_units is input_shape[0]
            # (3, 64, 64, 2) where 3 is input and 2 is latent space will become
            # (64, 64, 3) when reversing.
            dense_units = (config.input_shape[0],) + config.dense_config.dense_units

            self.network = DenseLayer(dense_units, config.activation, reverse=True)

        elif config.cnn_config and not config.dense_config:
            # spatial dimensions at the end of the CNN layer
            spatial_dims = [
                dim // (2 ** len(config.cnn_config.stack))
                for dim in config.input_shape[1:]
            ]

            # flatten output from the cnn which depends on dimensions of x/y and the final number
            # of filters.
            cnn_flat_dims = cnn_output_flat(
                len(config.cnn_config.stack),
                config.input_shape[1:],
                config.cnn_config.stack_filters[-1],
            )

            unflatten_shape = (
                cnn_flat_dims // (spatial_dims[0] * spatial_dims[1]),
                *spatial_dims,
            )

            # need to unflatten from latent space to correct cnn input dimension
            self.unflatten = nn.Unflatten(1, unflatten_shape)

            self.cnn = CNNLayer(
                config.cnn_config.stack,
                config.cnn_config.stack_filters,
                config.cnn_config.kernel_size,
                config.input_shape,
                config.activation,
                reverse=True,
            )

            self.network = nn.Sequential(self.unflatten, self.cnn)

        elif config.cnn_config and config.dense_config:
            ffn_input = cnn_output_flat(
                len(config.cnn_config.stack),
                config.input_shape[1:],
                config.cnn_config.stack_filters[-1],
            )

            # NOTE:
            # Will be reversed when buildind dense layers such that
            # last layer of FFN feeding into CNN has the right (flatten) output
            # dimension.
            dense_units = (ffn_input,) + config.dense_config.dense_units

            self.ffn = DenseLayer(dense_units, config.activation, reverse=True)

            spatial_dims = [
                dim // (2 ** len(config.cnn_config.stack))
                for dim in config.input_shape[1:]
            ]

            unflatten_shape = (config.cnn_config.stack_filters[-1], *spatial_dims)

            self.unflatten = nn.Unflatten(1, unflatten_shape)

            self.cnn = CNNLayer(
                config.cnn_config.stack,
                config.cnn_config.stack_filters,
                config.cnn_config.kernel_size,
                config.input_shape,
                config.activation,
                reverse=True,
            )

            self.network = nn.Sequential(self.ffn, self.unflatten, self.cnn)

    def forward(self, x):
        return self.network(x)


class Diresa(nn.Module):
    """
    Distance Regularized autoencoder.

    Build either using class method ``from_hyper_param`` which builds the encoder and
    decoder from a list of hyper parameters.
    Can also be built by providing custom encoder and decoder using either class method ``from_custom`` or
    by directly calling the constructor as Diresa(custom_encoder, custom_decoder).
    """

    @classmethod
    def from_hyper_param(
        cls,
        # Global parameters
        input_shape: Tuple[int, ...] = (),
        activation: nn.Module = nn.ReLU(),
        # Dense parameters
        dense_units: Optional[Tuple[int, ...]] = None,
        # CNN parameters
        stack: Optional[int] = None,
        stack_filters: Optional[Tuple[int, ...]] = None,
        kernel_size: Tuple[int, int] = (3, 3),
    ):
        """
        Builds DIRESA from hyperparameters

        :param input_shape: is the shape of input features.
        :param stack: are the number of Conv2D layers in each block
        :param stack_filters: number of filters in a block
        :param kernel_size: kernel size for convultion
        :param dense_units: tuple which describes the width of each feed forward layer. First element describes the first hidden layer. Last layer defines the output layer which is the size of latent space. E.g. ``(16, 32, 64)`` is a FFN with two hidden layers ``(16, 32)`` and last layer which defines latent space of size ``64``. When only one value is present it represents a linear projection on to a latent space with dimensions of that value. The Diresa class is responsible for adding the input layer to the FFN.
        :param activation: activation function used through the network.
        """

        # Create global configuration with sub-configs
        # General config is not saved at the moment in DIRESA class.
        # => Not possible to retrieve it by calling Diresa.config.
        config = DiresaSetup.create_config(
            stack=stack,
            stack_filters=stack_filters,
            kernel_size=kernel_size,
            input_shape=input_shape,
            dense_units=dense_units,
            activation=activation,
        )

        return cls(Encoder(config), Decoder(config))

    @classmethod
    def from_custom(cls, encoder: nn.Module, decoder: nn.Module):
        """
        Builds Diresa from custom provided encoder and decoder
        Identical to DIRESA(encoder, decoder)
        """
        return cls(encoder, decoder)

    # Could make constructor private to avoid direct instanciation.
    # Howver does not make a difference at the moment as there are
    # no additional checks in `from_custom`.

    def __init__(self, encoder: nn.Module, decoder: nn.Module):
        super().__init__()

        # Called base as those are the simple encoders/decoders which
        # produce latent space and decode latent space
        # but do not provide any additional features such as distance computation
        # or ordering
        self.base_encoder = encoder
        self.dist_layer: DistanceLayer = DistanceLayer(dim_less=True)
        self.base_decoder = decoder
        self.ordering_layer = OrderingLayer()

    def __random_ordering_index(self, x):
        """
        Shuffles batch x.
        """
        batch_size = x.size(0)
        perm_indices = torch.randperm(batch_size, device=x.device)
        return perm_indices

    def get_ordering(self):
        """
        :param y: original dataset
        :param y_pred: predicted dataset
        :return: R2 score between y and y_pred
        """
        return self.ordering_layer.order

    def is_ordered(self):
        """
        Returns True if ordering has been called on model.
        False otherwise.
        """
        return self.ordering_layer.order is not None

    def reset_ordering(self):
        """
        Removes current ordering and returns it.
        :return: Removed ordering
        """
        current = self.ordering_layer.order
        self.ordering_layer.order = None
        return current

    @staticmethod
    def __warn_ordering(f):
        """
        Decorator used to produce a warning when no ordering has been computed
        """

        def check_order(*args, **kwargs):
            self = args[0]
            if not self.is_ordered():
                # NOTE: Could improve doc here
                logging.warning(
                    "Latent components of Diresa have not yet been ordered. To produce an ordering of latent components call order_diresa() from trainer module"
                )
            return f(*args, **kwargs)

        return check_order

    @__warn_ordering
    def encode(self, x):
        """
        Runs encoder to provide latent space representation.
        :param x: batch of original data
        """
        encoded = self.base_encoder(x)
        ordered = self.ordering_layer(encoded)
        return ordered

    @__warn_ordering
    def decode(self, x):
        """
        Runs decoder to provide reconstructed data from latent space representation.
        :param x: batch of latent data.
        """
        reversed_ordering = self.ordering_layer(x, reverse=True)
        return self.base_decoder(reversed_ordering)

    # TODO: Could find a a naming scheme here to show
    # that this is part of the DIRESA arch.
    def _encode_with_distance(self, x):
        """
        Encodes batch `x` into latent space and compute distance between encoder
        and twin encoder in original and latent space.

        :param x: Batch to encode
        :return: (latent, distance) which are the encoded latent variables as well as
            the distance between the encoder and twin_encoder in original and latent space.
        """
        x1 = x
        y1 = self.base_encoder(x1)

        # Twin encoder
        index = self.__random_ordering_index(x)
        x2 = x1[index]
        y2 = y1[index]

        dist = self.dist_layer(x1, x2, y1, y2)
        return y1, dist

    # NOTE: We could have different approaches here
    # - Could either run the full model in train mode and restrict eval mode to
    #   only do encoder/decoder steps without computing distances and covariance.
    #   This enables faster inference but makes it impossible to evaluate distance
    #   and cov losses
    # - Or could use full model as well in eval mode. This would make it possible
    #   to get distance and cov loss on val/test sets. Then provide an additional
    #   method which skips internal layers and does simple encoder/decoder steps
    #   for inference purposes only.
    # Current approach: No difference between training and validation. Both use the
    # full model. Use trainer.predict function to bypass distance layer.
    #

    def forward(self, x):
        """
        Produces ``reconstructed``, ``latent`` and ``distance`` information
        """
        latent, dist = self._encode_with_distance(x)

        reconstructed = self.base_decoder(latent)

        # returns reconstrcuted for reconstruction loss
        # y1 is latent, used for covariance loss
        # returns distances between points for distance loss
        return reconstructed, latent, dist

    def fast_eval(self, x):
        """
        Produces an reconstructed output by using only encoder and decoder.
        This skips passes in the distance and covariance layers.
        """
        return self.base_decoder(self.base_encoder(x))
